/* GSI - general sound interface
 * gsiplay: play a sample/midi with the gsi server or set hardware volumes
 *
 * main.cc
 *
 * By W.H.Scholten 1998
 *
 * License for this file: Public domain, use as a basis for your own programs
 */


/* This program plays raw files (which MUST have the extension .raw, and are
 * assumed to be 8 bit mono 8000Hz), .au/.snd, .aif[fc], .wav, midi files,
 * and can set the master_volume, pcm_volume, synth_volume, and cd_volume.
 *
 *    gsiplay [ -v <softvol> ] [ file 1 ] [ file 2 ] ...
 *            [ -m <master> ] [ -p <pcm> ] [ -s <synth> ] [ -c <cd> ]
 *
 *    all options can be given multiple times like this:
 *      'gsiplay -v 128 file1 -v 200 file2'
 *    as all arguments are executed sequentially.
 *
 *    The supported file types are:
 *            .raw (raw; 8 bit unsigned mono assumed), .au/.snd (Sun/SGI),
 *            .aif[fc] (Apple audio interchange file format), .wav (yuk!)
 *            .mid (midi)
 *
 * When USE_CURSES is defined (see the code below) volume can be adjusted at 
 * runtime (when running in the foreground) by pressing -/+ (or shifted, i.e.
 * _/=)
 *
 * ?? add option: -speed <n> -bit <8/16> -mono/-stereo for raw samples ? 
 *
 * see pcmio options:
 *     pcmio +play,stereo,44100,s16,loop 50s:20s song
 * ?? should I do something like this?
 * ?? s8, 8000
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include <math.h>

#include <gsi/mytypes.h>
#include <gsi/debug.h>

#include <gsi/gsi.h>
#define USE_CURSES
#ifdef USE_CURSES
#  include <curses.h>
#endif

/*__________________DEFINITIES___________________*/
#define max(a,b) (a>b ?a:b)

/*__________________VARIABELEN___________________*/
int  snd_channels=1;
int  frag_size=8;
int  snd_type=GSI_8BIT;
int  dsp_speed=8000;

int  called_back=False;

/*__________________PROGRAMMA___________________*/
void my_callback(int ch) {
	DEB1(fprintf(stderr, "gsiplay: Callback from gsi_server (ch=%d)\n", ch));
	if (ch==0 || ch==-1) /* -1 =midi end callback */
		called_back=True;
}
void error_callback(int error) {
	DEB1(fprintf(stderr, "gsiplay: Error from gsi_server\n"));
	called_back=True;
}

void use() {
	fprintf(stderr,"  use: gsiplay [ -v <softvol> ] [ file ] [ file 2 ] ...\n"
                   "               [ -m <master> ] [ -p <pcm> ] [ -s <synth> ] [ -c <cd> ]\n"
                   "  notes: volumes are 0-256, a +/- before a volume changes it\n"
                   "         default softvolume is 128 (to avoid clipping)\n\n");
	exit(0);
}

#ifdef USE_CURSES
void close_curses() {
	nocbreak(); /* buffer IO again */
	endwin();
}
#endif

int main(int argc, char *argv[]) {
	int  i, j;
	int  softvolume=128; /* the software mixing volume */
	char pwd_buf[256];
	int  status;
	char *ptr;

	ptr=getcwd(pwd_buf, 256); /* longer and you're screwed */
	if (ptr==NULL) { perror("getcwd"); exit(1); }

	if (argc == 1) {
		use();
	}

#ifdef USE_CURSES
	// see man (n)curses for more info on these functions
	initscr();
	cbreak();
	//noecho();

	//nonl(); /* no newline */
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);

	atexit(close_curses);
#endif

	gsi_init();

	for (j=1; j<argc; j++) {
		if (argv[j][0]=='-') {
			if (argc<=j+1) use();
			else {
				int  vol_change=0;
				if (argv[j+1][0] =='+' || argv[j+1][0] =='-') vol_change++;

				switch(argv[j][1]) {
				case 'm': gsi_set_master_volume(atol(argv[j+1])); break;
				case 'p': gsi_set_pcm_volume(atol(argv[j+1])); break;
				case 's': gsi_set_synth_volume(atol(argv[j+1])); break;
				case 'c': gsi_set_cd_volume(atol(argv[j+1])); break;
				case 'v': softvolume=atol(argv[j+1]); break;
				default: use();
				}
				j++;
			}
			continue;
		}

		if (!strcmp(&argv[j][max(0, strlen(argv[1])-4)], ".mid")) { /* midi */
			gsi_init_synth (0);

			gsi_load_song_file (argv[1]);
			gsi_play_song (1);
		}
		else { /* samples */
			//.raw=assume 8 bit mono 8 kHz
			frag_size = 8+snd_channels-1+(snd_type+1)/2-1;
			gsi_init_pcm(dsp_speed, snd_channels, snd_type, frag_size, pwd_buf, NULL);
			gsi_send_commands(1, SND_CALLBACK_ON);
			gsi_send_commands(1, SND_USE_BEST_DSP_SETTINGS);
			gsi_load_file(0, argv[j]);
			gsi_set_channel_callback(0, my_callback);
			gsi_set_error_callback(error_callback);

			gsi_play_channel_sound(0, 0, SND_PLAY);
			gsi_send_commands(2, SND_SET_CHANNEL_VOLUME | SND_CURRENT, softvolume);
		}

#if 0
		close(0);
		close(1);
		if ((fdCon = open("/dev/tty", O_RDWR )) < 0) {
			perror("open of /dev/tty failed");
			return -1;
		}
#endif
		/* Now wait for the playing-done callback (or an error) */
		called_back=False;
		for (;;) {
			fd_set sel;
			int  selno;

			gsi_flush(); /* also handles the callbacks */

			if (called_back) break;
			FD_ZERO(&sel);
#ifdef USE_CURSES
			FD_SET(0, &sel); /* watch stdin for -/+ */
#endif
//			FD_SET(fdCon, &sel);
			FD_SET(gsi_callback_fd (), &sel);
			selno=gsi_callback_fd ()+1;
			select (selno, &sel, NULL, NULL, NULL);
#ifdef USE_CURSES
			if (FD_ISSET(0, &sel)) {
				byte b;
				b=getchar();
				if (b=='-' || b=='_') {
					softvolume -=5; if (softvolume<0) softvolume=0;
					gsi_send_commands(2, SND_SET_CHANNEL_VOLUME | SND_CURRENT, softvolume);
					DEB1(fprintf(stderr, "- : softvolume = %d\n", softvolume));
				}
				if (b=='=' || b=='+') {
					softvolume +=5; if (softvolume>255) softvolume=255;
					gsi_send_commands(2, SND_SET_CHANNEL_VOLUME | SND_CURRENT, softvolume);
					DEB1(fprintf(stderr, "+ : softvolume = %d\n", softvolume));
				}
			}
#endif /* USE_CURSES */
		}
		softvolume=128;
	} /* loop through all arguments */

	gsi_close();
	return 0;
}
