/* GSI - general sound interface
 *
 * gsi_midi.c (MIDI playing)
 *
 * Copyright (C) W.H.Scholten 1998
 *
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/* OpenBSD and NetBSD synth devices ? /dev/audio only at the moment.
 * OpenBSD: /dev/sound, /dev/audio, /dev/speaker
 * What are the ioctl's ? */

#include <limits.h>

#include <gsi/mytypes.h> /* use gsi_int16 etc instead ? */
#include <gsi/debug.h>
#include <gsi/config.h>
#include <gsi/gsi_commands.h>
#include "midi.h"
#include "gsi_server.h"

#ifdef USE_OSS
#ifdef __linux__
#  include <sys/soundcard.h>
#  include <linux/awe_voice.h>
#  define SEQUENCER_DEV "/dev/sequencer"
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
//#  include <machine/awe_voice.h> /* ?? */
#  define SEQUENCER_DEV "/dev/synth"
#endif
#endif


/*__________________DEFINITIES___________________*/
#define GSI_ERROR_MALLOC 10

#ifdef USE_OSS
#ifdef USE_OSS_38_FEATURES
#  if SOUND_VERSION < 0x030802
#    error You need OSS 3.8 or higher
#  endif
#endif
#endif

#ifndef SEQUENCER_DEV
#define SEQUENCER_DEV ""
#endif


/*__________________EXTERNALS___________________*/
extern int  playsong;

extern int  current_client;
extern struct client_data *current;


/*__________________VARIABELEN___________________*/
/* I currently allow playing a midi and ONE synth together, no more */
int  best_synth=-1;

int  seqfd = -1;

static int  synth_initialized = False;

#ifdef USE_OSS
SEQ_USE_EXTBUF();
#endif

int  nrsynths = 0;
int  *synth_types; /* niet meer nodig ? */
struct synth_info *card_info; /* One struct per soundcard */

char *seq_device = SEQUENCER_DEV;
int  chn_mask = 0xffff; /* only channels with a bit set are played, for interactive music programs */
int  perc_chn = PERCUSSION_CHANNELS; /* These channels play percussion, default = channel 10 */
/*use channel masks for each device ... ? */

int  playsong=0;
int  playsong_ct=0; /* -1 =keep playing, 0=finished, >0 not done (decrement by one each time the song finishes) */


int  ticks;
/*struct timeval start_time;*/

int  channel_output[16]; /* -1=external midi, >=0 is a device in card_info[] */

int  reverb, chorus;
float speed=1.0;


int  patchloaded[256];
int  fmloaded[256]; /* replace by int load_flags[8] and some other logic */
/* could also use byte table when translation tables are used (to show which patch was really loaded) NO: that can be extra */
/* Override volumes and pgm's (these will be reset when loading a MIDI or when a GSI_MIDI_RESET is given */
int  force_chn_volume[16]; /* value = volume+1 or 0 for no forced replacement */
int  force_chn_pgm[16]; /* value = pgm+1 or 0 for no forced replacement */

char *file_buf; /* the midi file */

int playing_song=1;
unsigned long default_tempo;

double current_time=0.0; /* MIDI timing */



/*__________________PROTOTYPES___________________*/
void preload_gus_patches ();
int  gsi_init_synth (int synth_type);

char *midi_device_type (int type);
char *synth_device_type (int type);
char *synth_device_subtype (int type);

int  load_song(char *name);
int  play_song(char *filename);

int  play_midi_init();
int  play_midi_events(int max_dt);

void gettimeofday_start();


/*__________________PROGRAMMA___________________*/
#ifdef SEPARATE
int main (int argc, char *argv[])
{
	if (argc<2) { printf("No file given\n"); exit(0); }

	play_song(argv[1]);
	return 0;
}


/*---------------------------------------------------------------------------*
 *	play!
 *---------------------------------------------------------------------------*/
int  play_song(char *filename)
{
	int  status;

	playing_song=0;
	status=load_song(filename); /* returns number of tracks */

	DEB1(fprintf(stderr, "read midi file: status=tracks=%d\n", status));

	if (status<=0) return -1;

	playing_song=1;
	play_midi_init();

	while (playing_song || _seqbufptr>0) play_midi_events(250000);

	close(seqfd);
	return 0;
}
#endif /* SEPARATE */


/*---------------------------------------------------------------------------*
 *	Reset file pointer etc
 *---------------------------------------------------------------------------*/
void reset_song() {
}


/*---------------------------------------------------------------------------*
 *	Timing
 *---------------------------------------------------------------------------*/
struct timeval start_tv;
void gettimeofday_start() {
	gettimeofday(&start_tv, NULL);
}

double gettimeofday_diff() {
	struct timeval new_tv;
	gettimeofday(&new_tv, NULL);
/*
	if (new_tv.tv_usec <start_tv.tv_usec) {
	}
	*/
	return (double) (new_tv.tv_sec-start_tv.tv_sec) + (double) (new_tv.tv_usec-start_tv.tv_usec) /1.0e6;
}


/*---------------------------------------------------------------------------*
 *	Returns number of tracks
 *---------------------------------------------------------------------------*/
int load_song(char *filename) {
	struct stat file_info;
	int  status=0;

	/* Load midi file */
	if (stat(filename, &file_info) != -1) {
		int midi_fd;
		DEB1(fprintf(stderr, "midi file: %d bytes\n", (int) file_info.st_size));
		if ((file_buf = malloc(file_info.st_size)) == NULL)
			return -GSI_ERROR_MALLOC;
		DEB1(fprintf(stderr, "reading midi file\n"));
		midi_fd=open(filename, O_RDONLY);
		read(midi_fd, file_buf, file_info.st_size);
		close(midi_fd);
		status = midi_to_tracks(file_buf, file_info.st_size);
	}
	return status; /* returns number of tracks */
}


int  format, numtracks, division;
byte *midifilebuf;
struct miditrack *song_tracks;

int16 msb_int16_at(byte *buf) {
	return (((int16) buf[0]) <<8) | buf[1];
}
int32 msb_int32_at(byte *buf) {
	return (((int32) buf[0]) <<24) | (((int32) buf[1]) <<16) | (((int32) buf[2]) <<8) | buf[3];
}

int read_msb_int32() {
	int a=msb_int32_at(midifilebuf);
	midifilebuf +=4;
	return a;
}
int read_msb_int16() {
	int a=msb_int16_at(midifilebuf);
	midifilebuf +=2;
	return a;
}

int midi_to_tracks(byte *filebuf, int filelength)
{
	unsigned int  i = 0;
	int  track, tracklen;

	midifilebuf = filebuf;

	i=read_msb_int32();

	if (i == MThd) {
		tracklen = read_msb_int32();
		DEB1(fprt(stderr, "tracklen=%d\n", tracklen));
		format = read_msb_int16();
		DEB1(fprt(stderr, "format=%d\n", format));
		numtracks = read_msb_int16();
		DEB1(fprt(stderr, "num_tracks=%d\n", numtracks));
		division = read_msb_int16();
		DEB1(fprt(stderr, "division=%d\n", division));
	} else {
		fprintf(stderr, "GSI: midi_to_tracks was given a bad midi file\n");
		return -1;
	}

	if (song_tracks !=NULL) free(song_tracks);
	song_tracks = malloc(sizeof(struct miditrack) *numtracks);

	for (track = 0; track < numtracks; track++) {
		if (read_msb_int32() != MTrk) {
			fprintf(stderr, "MTrk not in the right place\n");
			exit(0);
			midifilebuf -= 3; /* go to next byte (note: read advances 4 bytes) */
			while (read_msb_int32() != MTrk && (midifilebuf < filebuf + filelength-8)) midifilebuf -= 3;
		}

		tracklen = read_msb_int32();

		/* Sanity check */
		tracklen = min(tracklen, filebuf + filelength - midifilebuf);

		song_tracks[track].length = tracklen;
		song_tracks[track].data = midifilebuf;
		midifilebuf += tracklen;
	}
	numtracks = track;

	return numtracks;
}


/*---------------------------------------------------------------------------*
 *	LOAD PATCHES FOR INSTRUMENTS IN THE FILE
 *	looks in the current midi (tracks) preloads all necessary patches on
 *	the GUS, need some runtime loading stuff if there's more than the
 *	available memory (add checks)
 *---------------------------------------------------------------------------*/
#ifdef USE_GUS_SYNTH
void preload_gus_patches() {
	int  i, j;


	DEB1(fprintf(stderr, "__GSI: loading gus patches\n"));

	DEB1(
		for (i = 0; i < numtracks; i++) {
			fprintf(stderr, "___GSI: track %d has length %d\n", i, (int) song_tracks[i].length);
		}
		);
	for (i = 0; i < numtracks; i++) {
		DEB1(fprintf(stderr, "___GSI: track %d\n", i));
		for (j = 0; j < (int) song_tracks[i].length - 5; j++) {
			int  chn = (song_tracks[i].data[j] & 0x0f);
			int  cmd = (song_tracks[i].data[j] & 0xf0);
			int  pgm = (song_tracks[i].data[j + 1]);

			if ((IS_GUS(chn)) && !(pgm & 0x80)) {
				if ((cmd == MIDI_PGM_CHANGE && !IS_PERC(chn))
					|| (cmd == MIDI_NOTEON && IS_PERC(chn))) {
					int patchno = pgm;

					if (IS_PERC(chn)) patchno |=128;
					else if (force_chn_pgm[chn]) patchno=force_chn_pgm[chn]-1;

					gus_load(seqfd, 0 /*type*/, 0 /*device*/, patchno /*pgm*/);
					DEB1(fprintf(stderr, "____GSI: CHN=%d, IS_PERC(chn)=%d, force_chn_pgm[chn]=%d\n\n", chn, IS_PERC(chn), force_chn_pgm[chn]));
				}
			}
		}
	}
}
#endif /* USE_GUS_SYNTH */


/*---------------------------------------------------------------------------*
 *	Send the sequencer data to the sound card
 *---------------------------------------------------------------------------*/
#ifdef USE_OSS /* This code is currrently OSS only */

#define WRITE_FD_V1
#ifdef WRITE_FD_V2
int _seqbeginptr=0;
void seqbuf_dump ()
{
	int status;
	DEB1(fprt(stderr, "_seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));

	if (_seqbufptr) {
		status= write (seqfd, ((char*) _seqbuf)+_seqbeginptr, (_seqbufptr-_seqbeginptr));
		DEB1(fprintf(stderr, "status=%d (bytes written)\n", status));

		if (status >0) {
			_seqbeginptr +=status;
			DEB1(fprt(stderr, "_seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));
		}
		if (_seqbeginptr==_seqbufptr) { _seqbufptr=0; _seqbeginptr=0; }
	}
	DEB1(fprt(stderr, "Reset _seqbufptr\n"));
}

void seqbuf_dump_soft() {
}
#endif
#ifdef WRITE_FD_V1
void seqbuf_dump ()
{
	int status;
	DEB1(fprt(stderr, "_seqbufptr=%d\n", _seqbufptr));

	if (_seqbufptr) {
		status= write (seqfd, _seqbuf, _seqbufptr);
		DEB1(fprintf(stderr, "status=%d (bytes written)\n", status));

		if (status>0) {
			memmove(_seqbuf, _seqbuf+status, _seqbufptr-status); 
			_seqbufptr -=status;
		}
	}
	DEB1(fprt(stderr, "Reset _seqbufptr\n"));
}
#endif
#ifdef WRITE_FD_V0
void seqbuf_dump ()
{
	int status, tmp;
	DEB1(fprt(stderr, "_seqbufptr=%d\n", _seqbufptr));

//status=(ioctl(seqfd, SNDCTL_SEQ_GETOUTCOUNT, &tmp)==-1);
//DEB1(fprintf(stderr, "status=%d, tmp=%d\n", status, tmp));
	if (_seqbufptr)
		if (write (seqfd, _seqbuf, _seqbufptr) == -1)
		{
			perror ("write /dev/sequencer");
			exit (-1);
		}
	_seqbufptr = 0;
	DEB1(fprt(stderr, "Reset _seqbufptr\n"));
}
#endif
#endif /* USE_OSS */


/*---------------------------------------------------------------------------*
 *	Initialise all synthesizer devices found 
 *---------------------------------------------------------------------------*/
int gsi_init_synth (int synth_type) {
#ifdef USE_OSS
	int  i;
	int  synth;
	int  status;
	int  best_synth_device=-2;

	DEB1(fprt(stderr, "__GSI: gsi_init_synth, type=%d\n", synth_type));

	if (ioctl(seqfd, SNDCTL_SEQ_NRSYNTHS, &nrsynths)==-1)
	{
		perror("GSI: unable to determine number of synthesizers");
		return -1;
	}

	DEB1(fprintf(stderr, "%s: %d synthesizer device%s installed\n", seq_device, nrsynths, nrsynths == 1 ? "" : "s"));

	if (nrsynths>0) {
		synth_types = calloc(nrsynths, sizeof(int));
		card_info = (struct synth_info *) malloc(nrsynths*sizeof(struct synth_info));
		voices = calloc(nrsynths, sizeof(void *));
		/* at exit: free(synth_types); free(card_info); free(voices[0-nrsynths-1]), free(voices); */

		/* Init all synth devices we can find */
		for (synth=0; synth < nrsynths; synth++)
		{
			card_info[synth].device = synth; /* overbodig? */
#define USE_OSS_350_FEATURES

#ifndef USE_OSS_350_FEATURES /* i.e. OSS >= 3.8 */
			status=ioctl(seqfd, SNDCTL_SYNTH_ID, &card_info[synth]);
#else
			status=ioctl(seqfd, SNDCTL_SYNTH_INFO, &card_info[synth]);
#endif
			if (status==-1) {
				perror("/dev/sequencer SYNTH_ID");
				return -1;
			}
			DEB1(fprt(stderr, "Device %d: `%s' type: `%s' subtype: `%s' voices: %d\n",
					 synth, card_info[synth].name, synth_device_type(card_info[synth].synth_type),
					 synth_device_subtype(card_info[synth].synth_subtype), card_info[synth].nr_voices
				));

			synth_types[synth] = SYNTH_HW_NONE;
#ifndef USE_OSS_350_FEATURES /* i.e. OSS >= 3.8 */
			if (strcmp(card_info[synth].name, "GUS")==0
				|| strcmp(card_info[synth].name, "IWAVE")==0
				|| strcmp(card_info[synth].name, "SoftOSS")==0)
#else
			if (card_info[synth].synth_type==SYNTH_TYPE_SAMPLE && card_info[synth].synth_subtype == SAMPLE_TYPE_GUS)
#endif /* USE_OSS_350_FEATURES n/y */
			{
#ifdef USE_GUS_SYNTH
				DEB1(fprt(stderr, "Found a GUS\n"));
				if (!gus_init(seqfd, synth)) {
					synth_types[synth] = SYNTH_HW_GUS;
					if (SYNTH_HW_GUS > best_synth_device) { best_synth_device=SYNTH_HW_GUS; best_synth=synth; }
				}
#endif
			}
#ifndef USE_OSS_350_FEATURES /* i.e. OSS >= 3.8 */
			else if (strcmp(card_info[synth].name, "OPL")==0)
#else
			else if (card_info[synth].synth_type==SYNTH_TYPE_FM && card_info[synth].synth_subtype == FM_TYPE_OPL3)
#endif /* USE_OSS_350_FEATURES n/y */
			{
#ifdef USE_OPL3_SYNTH
				DEB1(fprt(stderr, "Found a OPL\n"));
				if (!opl3_init(seqfd, synth)) {
					synth_types[synth] = SYNTH_HW_OPL;
					if (SYNTH_HW_OPL > best_synth_device) { best_synth_device=SYNTH_HW_OPL; best_synth=synth; }
				}
#endif
			}
#ifndef USE_OSS_350_FEATURES /* i.e. OSS >= 3.8 */
			/* What's the string for a AWE32? */
			else if (strcmp(card_info[synth].name, "AWE32")==0)
			{
#ifdef USE_AWE_SYNTH
				/* How to init awe32? */
//				voices[synth] = malloc(sizeof(voice_state) * card_info[dev].nr_voices);
#endif
			}
#else
#  undef USE_AWE_SYNTH
#endif /* USE_OSS_350_FEATURES n/y */
		}
		synth_initialized = True;
	}
	DEB1(fprintf(stderr, "++: best synth = %d\n", best_synth));
	for (i=0; i<16; i++) { channel_output[i]=best_synth; } /* default */
#endif /* USE_OSS */

	return 0;
}


/*---------------------------------------------------------------------------*
 *	Debug and information stuff
 *---------------------------------------------------------------------------*/
#ifdef USE_OSS
DEB(
char *synth_device_type(int type)
{
	switch (type) {
	case SYNTH_TYPE_FM:     return("FM");
	case SYNTH_TYPE_SAMPLE: return("Sampling");
	case SYNTH_TYPE_MIDI:   return("MIDI");
	default:                return("unknown");
	}
}

char *synth_device_subtype(int type)
{
	switch (type) {
	case FM_TYPE_ADLIB:   return("Adlib");
	case FM_TYPE_OPL3:    return("OPL-3");
	case SAMPLE_TYPE_GUS: return("GUS");
	default:              return("unknown");
	}
}
);
#endif

/*---------------------------------------------------------------------------*
 *	Patch loading
 *---------------------------------------------------------------------------*/
#if 0
/* Haven't finished this version */
int patch_len;

int load_patchfile(int synth_dev, int pgm) {
	byte *patch;
/*	int patch_len;*/

	switch (synth_types[synth_dev]) {
	case SYNTH_HW_GUS:
		patch_len=gsi_open_patch(pgm, patch);
		if ( patch_len>0 ) {
			gus_load(synth_dev, patch, patch_len);
			gsi_close_patch();
		}
		break;
	case SYNTH_HW_OPL:
		if (!gsi_open_patch(pgm, patch) ) {
			patch_len=20;
			opl3_load(synth_dev, patch, patch_len);
			gsi_close_patch();
		}
		break;
	default: fprintf(stderr, "unknown synth device %d\n", synth_types[synth_dev]);
	}
}

int load_patch(int synth_dev, int patch_len, byte *patch) {
	switch (synth_types[synth_dev]) {
	case SYNTH_HW_GUS:
		gus_load(synth_dev, patch, patch_len);
		break;
	case SYNTH_HW_OPL:
		opl3_load(synth_dev, patch, patch_len);
		break;
	default: fprintf(stderr, "unknown synth device %d\n", synth_types[synth_dev]);
	}
}
#endif /* 0 */


/*---------------------------------------------------------------------------*
 * Read variable-length number (7 bits per byte, MSB first)
 *---------------------------------------------------------------------------*/
unsigned long read_vl(struct miditrack *s)
{
	unsigned long len = 0;
	byte c;

	for (;;) {
		if (s->index >= s->length) break; /* sanity check */
		c = s->data[(s->index)++];
		len = (len<<7) | (c & 0x7f);
		if (!(c & 0x80)) break;
	}
	return len;
}


/* the length of MIDI commands: */
int cmd_len[16] ={
	0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 1, 1, 2, 0
};
/* Midi commands are:
 * 0-7: misc.
 * enum { NOTE_OFF = 8, NOTE_ON, AFTER_TOUCH, CONTROL_CHANGE, PROGRAM_CHANGE,
 *    CHANNEL_PRESSURE, PITCH_WHEEL, SYSEX };
 * (after touch=key pressure, pitch_wheel=pitch_bend SYSEX=MIDI_SYSTEM_PREFIX)
 */


/*---------------------------------------------------------------------------*
 *	Play the midi events on internal sythesizers or external midi
 *	Initialization
 *---------------------------------------------------------------------------*/
unsigned long tempo;
unsigned long lasttime = 0;
unsigned int  earliest, track, best_track, length;
unsigned char *data;
double current_midi = 0.0;
double dtime = 0.0;
double last_midi_write = 0.0;

int play_midi_init()
{
	int  track, chan, dev;
	int  k;

	/* try to open seq_device */
	seqfd = open(seq_device, O_WRONLY);
	if (seqfd == -1) {
		fprintf(stderr, "GSI: unable to open `%s', ", seq_device);
		return 1;
	}

	/* Non blocking writes */
#define NON_BLOCKING_SEQ_WRITE
#ifdef NON_BLOCKING_SEQ_WRITE
	fcntl(seqfd, F_SETFL, O_NONBLOCK);
#endif

	gsi_init_synth(0 /*synth_type*/);

	default_tempo = 500000;

	/* Actual midi playing: */
	for (k = 0; k < 16; k++)
		force_chn_pgm[k] = force_chn_volume[k] = 0;	/* no pgm/vol override */

#ifdef USE_GUS_SYNTH
	preload_gus_patches();
#endif

#ifdef USE_OSS
	seq_reset();
#endif

/*	gettimeofday(&start_time, NULL);*/	/* to sync */
	gettimeofday_start();

	for (track = 0; track < numtracks; track++) {
		song_tracks[track].index = song_tracks[track].running = 0;
		song_tracks[track].ticks = read_vl(&song_tracks[track]);
	}
#ifdef USE_OSS
	for (chan = 0; chan < 16; chan++) {
		dev=channel_output[chan];
		DEB1(fprintf(stderr, "++: dev=%d\n", dev));
		seq_control(dev, chan, CTL_BANK_SELECT, 0);
		seq_control(dev, chan, CTL_EXT_EFF_DEPTH, reverb);
		seq_control(dev, chan, CTL_CHORUS_DEPTH, chorus);
		seq_control(dev, chan, CTL_MAIN_VOLUME, 127);
		seq_chn_pressure(dev, chan, 127);
	}
	SEQ_START_TIMER();
	SEQ_DUMPBUF();
#endif

	tempo = default_tempo;
	lasttime = 0;

	current_midi = 0.0;
	last_midi_write = current_midi; /* for DSP write etc */
	dtime = 0.0; /* time diff */

	return 0;
}


/*---------------------------------------------------------------------------*
 *	Play the midi events on internal sythesizers or external midi
 *	Main loop
 *---------------------------------------------------------------------------*/
/* Some often used abbreviations */
#define TIME	song_tracks[track].ticks
#define CMD		song_tracks[track].running
#define CHN		(CMD & 0xf)
#define NOTE	data[0]
#define VEL		data[1]

int play_midi_events(int max_dt)
{
#ifdef USE_OSS
	int  patch_loaded;
	int  use_dev;

	if (_seqbufptr<=400) {
		for(;;) {
			if (_seqbufptr>400) { SEQ_DUMPBUF(); break; }

			/* Rename: best_track -> earliest_track / first_track */
		earliest = UINT_MAX;
		for (best_track = track = 0; track < numtracks; track++) {
			if (song_tracks[track].ticks < earliest) {
				best_track = track;
				earliest = /*TIME*/ song_tracks[track].ticks;
			}
		}
		if (earliest == UINT_MAX) {
			playing_song=0;
			return 1;		/* no more data to read */
		}

		/* Take the earliest time, if it's later than
		 *   max(0.4s, 4* fragtime)
		 * [ where fragtime=fragsize/(dsp_chan*dsp_bytes*dsp_speed) ]
		 * then leave and do it next time as we use the blocking dsp write
		 * to time and may want to pause midi playing
		 */
		/* 102.4 = 1 s */
		/* max_dt in musec */
		DEB(fprintf(stderr, "t=%f cur=%f, last=%f, last+dt=%f\n", ((earliest * (tempo / 10000)) / (double) (division)) * speed, current_midi, last_midi_write, last_midi_write+(double) (102*max_dt)/1.0e6));
#if 1
		if ( ((earliest * (tempo / 10000)) / (double) (division)) * speed > last_midi_write+(double) (102*max_dt)/1.0e6) {
			break;
		}
#endif

		track = best_track;

		use_dev=channel_output[CHN];
		DEB1(fprintf(stderr, "use_dev=%d\n", use_dev));

		{
		byte *trk_data = song_tracks[track].data;
		unsigned long index = song_tracks[track].index;
		unsigned long trk_length = song_tracks[track].length;
		byte cmd = song_tracks[track].running;
		unsigned long time = song_tracks[track].ticks;

		/* Look at the midi file data for instructions */
		if ((trk_data[index] & 0x80) && (index < trk_length)) {
			cmd = trk_data[index];
			index++;
		}

		/* Do some preliminary decoding and skipping of unhandled stuff */
		if (cmd == 0xff && index < trk_length)
			cmd = trk_data[index++]; /* next instruction */

		if (cmd > 0xf7)	{
			/* midi real-time message (currently ignored)
			 * (will come in next version) */
			length = 0;
		}
		else {
			length = cmd_len[cmd >> 4];
			if (length==0) {
				song_tracks[track].index=index;
				length = read_vl(&song_tracks[track]);
				index=song_tracks[track].index;
			}
		}

		if (index + length < trk_length) {
			data = &(trk_data[index]);
			if (cmd == MIDI_Set_tempo)
				tempo = ((data[0] << 16) | (data[1] << 8) | data[2]);
			if (time > lasttime) {
				if (division > 0) {
					dtime = ( (double) ((time - lasttime) * (tempo / 10000)) / (double) (division)) * speed;
					current_midi += dtime;
					lasttime = time;
				} else if (division < 0) {
					current_midi = ( (double) time / ( (double) ((division & 0xff00 >> 8) * (division & 0xff)) * 10000.0) ) * speed;
				}

				/* sanity check: stop if for 20 seconds or more there's
				 * nothing to be played (nicely rules out that insane
				 * silence piece) */
				if (dtime > 2048.0)
					playing_song = 0;
				else if ((int) current_midi > ticks) {
					SEQ_WAIT_TIME((ticks = (int) current_midi));
					SEQ_DUMPBUF();
					/* replace by a seq_dumpbuf_soft(); which doesn't
					 * flush all data, which may be needed for the
					 * macrodefinitions that store in the buffer */
				}
			}
			song_tracks[track].running=cmd;
			song_tracks[track].ticks=time;
			song_tracks[track].index=index;

			patch_loaded = 0;

			if (playing_song && CMD > 0x7f && IS_PLAYING(CHN))
				switch (CMD & 0xf0) {
				case MIDI_NOTEOFF:
					seq_stop_note(use_dev, CHN, NOTE, VEL);
					break;
				case MIDI_NOTEON:
					if (IS_PERC(CHN) && VEL && (!IS_MIDI(CHN)) )
						patch_loaded = seq_set_patch(CHN, NOTE + 128);
					if (VEL && force_chn_volume[CHN])
						VEL = force_chn_volume[CHN]-1;
					if (patch_loaded != -1)
						seq_start_note(use_dev, CHN, NOTE, VEL);
					break;
				case MIDI_KEY_PRESSURE:
					if (IS_PERC(CHN) && VEL && (!IS_MIDI(CHN)))
						patch_loaded = seq_set_patch(CHN, NOTE + 128);
					if (patch_loaded != -1)
						seq_key_pressure(use_dev, CHN, NOTE, VEL);
					break;
				case MIDI_CTL_CHANGE:
					seq_control(use_dev, CHN, NOTE, VEL);
					break;
				case MIDI_PGM_CHANGE:
					if (IS_MIDI(CHN) || !IS_PERC(CHN))
						NOTE = seq_set_patch(CHN, NOTE);
					break;
				case MIDI_CHN_PRESSURE:
					seq_chn_pressure(use_dev, CHN, NOTE);
					break;
				case MIDI_PITCH_BEND:
					seq_bender(use_dev, CHN, NOTE, VEL);
					break;
				case MIDI_SYSTEM_PREFIX:
					/* The sysex decoding is very incomplete at the moment,
					 * should be better in GSI 0.6 */
					if (length > 1) {
						unsigned long i, j;

						DEB1(fprt(stderr, "__GSI: decode_sysex\n"));
						if (length > 7 && data[0] == 0x41 && data[2] == 0x42 && data[3] == 0x12) {
							/* GS data set messages */
							if (data[4] == 0x40 && (data[5] & 0xf0) == 0x10 && data[6] == 0x15) {
								/* Use rythm part (rythm channels used for drums) */
								if (!(i = (data[5] & 0xf)))
									i = 0x09;
								else if (i < 10)
									i--;
								i = 1<<i;
								if (data[7])
									perc_chn |= i;
								else
									perc_chn &= ~i;
							}
							if ((data[4] == 0x40 || data[4] == 0) &&
								data[5] == 0x00 && data[6] == 0x7f) {
								/* GS reset */
								perc_chn = 0x0200;	/* percussion only in channel 10 (no more rythm channels for drums) */
								for (i = 0; i < 16; i++) {	/* set state info */
									for (j = 0; j < 128; j++)
										note_vel[i][j] = 0;
									channel[i].bender = 8192;
									channel[i].bender_range = 2;
									channel[i].controller[CTL_PAN] = 64;
									channel[i].controller[CTL_SUSTAIN] = 0;
								}
							}
							/* All this handles the first of alienz' sysex messages, but the second, hmm. */
						}
						/* Send sysex to Midi out */
						for (j=0; j<nrsynths; j++) {
							if (synth_types[j]==MIDI_HW) {
								if (/*type*/ CMD == MIDI_SYSTEM_PREFIX)
									SEQ_MIDIOUT(j, MIDI_SYSTEM_PREFIX);
								for (i = 0; i < length; i++)
									SEQ_MIDIOUT(j, data[i]);
							}
						}
					}
					break;
				default:
					break;
				} /* switch (cmd &0xf0) */

#if 0
			/* Not yet implemented: */
			/* To send back data to the main program (along with callback
			 * data) [prefix: SND_MIDI_EVENT] and sync.
			 * This is mainly of interest to graphical MIDI players */
			if (midi_info_feedback) {
				send_midi_status(CMD, data, length);
				ioctl(seqfd, SNDCTL_SEQ_SYNC); /* necessary? */
			}
#endif
		}

		/* adjust MIDI time */
		song_tracks[track].index += length;
		if (song_tracks[track].index >= song_tracks[track].length)
			song_tracks[track].ticks = UINT_MAX;	/* track done */
		else
			song_tracks[track].ticks += read_vl(&song_tracks[track]);

	}
		} /* for(;;) */
	}
	else { SEQ_DUMPBUF(); }

	if (playing_song==0 && _seqbufptr==0) {
			/* We're done */
		if (playsong_ct>0) playsong_ct--;
		if (!playsong_ct) playsong=0; else playing_song=1;
		playsong=0; /* flag in the main server code */
#ifndef SEPARATE
		if (current->do_callback) {
			status_send_byte(GSI_CALLBACK_SONG_END);
			DEB1(fprintf(stderr, "GSI_SERVER: callback in queue\n"));
		}
#endif
	}

#endif /* USE_OSS */
	return 1;
}

/*---------------------------------------------------------------------------*
 *	EOF gsi_midi.c
 *---------------------------------------------------------------------------*/
